// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.3.0
//                        (increment this number EVERY time you make a change)
//
// 2.3.0        26 Jun 1997     Jill Baker
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

/*
 * Copyright (C) 1995, 1996 Systemics Ltd (http://www.systemics.com/)
 * All rights reserved.
 *
 * This library and applications are FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 * as long as the conditions within the COPYRIGHT file are adhered to.
 *
 */

package cryptix.io;

import cryptix.io.Lang;
import cryptix.pgp.charset.CharSet;
import cryptix.pgp.FormatException;
import cryptix.security.MD5;
import cryptix.util.SensitiveByteArray;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

/**
 * @author Jill Baker
 */
public abstract class UserInterface
{
    public final static String yes              = Lang.translate( "Yes" );
    public final static String no               = Lang.translate( "No" );
    public final static String y                = Lang.translate( "y" );
    public final static String n                = Lang.translate( "n" );
    public final static String passphrasePrompt = Lang.translate( "Enter passphrase" );

    private MD5 passphraseMD5 = new MD5();
    private ByteArrayOutputStream entropyPool = new ByteArrayOutputStream();
    private int entropySize = 0;
    private int passphraseLength = 0;

    protected boolean verbose = false;
    public boolean getVerbose(){ return verbose; }
    public void setVerbose( boolean b ){ verbose = b; }

    protected boolean showPassphrase = false;
    public boolean getShowPassphrase(){ return showPassphrase; }
    public void setShowPassphrase( boolean b ){ showPassphrase = b; }

    /**
	 * Display a status message
	 * @param message the message to be displayed
	 */
	public abstract void status( String message );

	/**
	 * Clear any existing status message
	 */
	public abstract void status();

	/**
	 * Notify the user of something
	 * @param message the message to be displayed
	 */
	public abstract void notice( String message );

	/**
	 * Warn the user of something
	 * @param message the message to be displayed
	 */
	public abstract void warning( String message );

	/**
	 * Inform the user of an error
	 * @param message the message to be displayed
	 */
	public abstract void error( String message );

	/**
	 * Prompt the user for a yes or no answer, and specify a default
	 * @param prompt the question to be displayed
	 * @returns true if yes, false if no
	 */
	public abstract boolean promptYesNo( String prompt, boolean def );

	/**
	 * Prompt the user for one of two choices
	 * @param prompt the question to be displayed
     * @param posText text for the positive button
     * @param negText text for the negative button
     * @param def default value
	 * @returns true if the positive button was pressed, false if the negative button was pressed.
	 */
	public abstract boolean promptBoolean( String prompt, String posText, String negText, boolean def );

    /**
     * Prompt the user for a String.
	 * @param prompt the question to be displayed
	 * @returns the String input.
     */
    public abstract String promptString( String prompt );

    /**
     * Prompt the user for a filename.
	 * @param prompt the question to be displayed
     * @param save true if file is to be saved, false if it is to be loaded
	 * @returns the file selected. (This is not guaranteed to exist).
     */
    public abstract File promptFilename( String prompt, boolean save );

    /**
     * Prompt the user for a selection from a list.
	 * @param prompt the question to be displayed
     * @param list Vector of Strings to be selected from
     * @param allowOthers true if user is allowed to type in entries not in the list
	 * @returns the file selected. (This is not guaranteed to exist).
     */
    public abstract String promptFromList( String prompt, Vector list, boolean allowOthers );

	/**
	 * Prompt the user for a passphrase.
     * @charset charset for character conversion
	 * @returns the MD5 hash of the passphrase, or null if cancel was selected
	 */
    public SensitiveByteArray promptPassphrase( String charset )
    {
        return promptPassphrase( passphrasePrompt, charset );
    }

	/**
	 * Prompt the user for a passphrase.
	 * @param prompt the prompt to be displayed
     * @charset charset for character conversion
	 * @returns the MD5 hash of the passphrase, or null if cancel was selected
	 */
    public SensitiveByteArray promptPassphrase( String prompt, String charset )
    {
        getPassphrase( prompt, charset );
        if (passphraseLength == 0) return null;
        return new SensitiveByteArray( passphraseMD5.digest(), false );
    }

	/**
	 * Prompt the user for a passphrase.
     * The child routine must call addPassphraseChar( ch, charset ) for each character in the passphrase.
	 * @param prompt the prompt to be displayed
     * @charset charset for character conversion
	 */
    protected abstract void getPassphrase( String prompt, String charset );

    /**
     * If necessary, prompt the user for some entropy
     */
    public byte[] promptEntropy( int numBytes )
    {
        while (numBytes > entropySize) getEntropy( numBytes - entropySize );
        byte[] entropyByteArray = entropyPool.toByteArray();
        byte[] returnByteArray = new byte[ numBytes ];
        System.arraycopy( entropyByteArray, 0, returnByteArray, 0, numBytes );
        entropyPool = new ByteArrayOutputStream();
        entropyPool.write( entropyByteArray, numBytes, entropyByteArray.length-numBytes );
        return entropyByteArray;
    }

    /**
     * Prompt the user for some entropy
     */
    protected abstract void getEntropy( int numBytes );

    /**
     * Add one character to the passphrase.
     * @param ch the character to add
     * @param charset the charset conversion to use
     */
    protected void addPassphraseChar( int ch, String charset )
    {
        byte[] b = new byte[1];
        try
        {
            b[0] = CharSet.decode( (byte) ch, charset );
        }
        catch (FormatException e)
        {
            b[0] = (byte) ch;
        }
        passphraseMD5.add( b );
        passphraseLength++;
    }

    /**
     * Add in some entropy.
     * @param entropy byte array containing some entropy
     */
    protected void addEntropy( byte[] entropy )
    {
        try
        {
            entropyPool.write( entropy );
            entropySize += entropy.length;
        }
        catch (IOException e){}
    }

    /**
     * Add in some entropy.
     * @param entropy long variable containing some entropy
     */
    protected void addEntropy( long entropy )
    {
        while ( entropy != 0 )
        {
            entropyPool.write( ((int) entropy) & 0xFF );
            entropy >>>= 8;
            entropySize++;
        }
    }
}
