// This file is currently locked by jbaker@madge.com
//
// Source code version is 2.5.6
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.6        04 Jun 1997     Jill Baker
//      State machine added so we can do line-at-a-time decrypting.
//      For this purpose I have added a constructor (boolean) and methods
//      add(), addLines() and isComplete(). I have also deprecated parse();
//
// 2.5.5        30 May 1997     Jill Baker
//      Added method capitalise(). Properties table now stored in mixed case -
//      Key _must_ have first char upper case, rest lower case. Bug fixed in
//      toString() in clearsigned message case.
//
// 2.5.4        29 May 1997     Jill Baker
//      Machine-specific linebreaks now dealt with correctly. Added new methods
//      dashProtect(), dashUnprotect() and removeTrailingSpaces(). Added
//      constructors (Properties, String, SecretKey) and (String, SecretKey).
//      Deprecated constructors (Properties, byte[], PacketByteArray) and
//      (byte[], PacketByteArray).
//
// 2.5.3        17 Apr 1997     Jill Baker
//      Changed imports to new namespace.
//
// 2.5.2        17 Apr 1997     Jill Baker
//      Version information now calculated, not fixed.
//
// 2.5.1        03 Apr 1997     Jill Baker
//      Added "Version: " information.
//      Quoted Printable now dealt with correctly.
//      Invalid header lines now throw a FormatException.
//      Header names now case insensitive.
//
// 2.5.0        25 Mar 1997     Jill Baker
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

/*
 * Copyright (C) 1995, 1996 Systemics Ltd (http://www.systemics.com/)
 * All rights reserved.
 *
 * This library and applications are FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 * as long as the conditions within the COPYRIGHT file are adhered to.
 *
 */

package cryptix.pgp;

import cryptix.mime.*;
import cryptix.security.Cryptix;
import cryptix.pgp.charset.CharSet;
import cryptix.security.MD5;
import cryptix.security.rsa.SecretKey;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * This class represents an Ascii-Armoured message.<p>
 *
 * @author Jill Baker
 * @author Ian Brown
 * @author Gary Howland
 *
 * @see cryptix.pgp.Signature
 * @see cryptix.pgp.app.ArmourDecryptor
 */
public final class ArmouredData
{
    // State machine states
    private final static int WAITING_FOR_BEGIN = 1;
    private final static int BEGIN_CLEARSIGNED_MESSAGE = 2;
    private final static int COMPILING_CLEARSIGNED_MESSAGE = 3;
    private final static int COMPILING_PROPERTIES = 4;
    private final static int DEARMOURING = 5;
    private final static int EXPECTING_END_OF_MESSAGE = 6;
    private final static int COMPLETE = 7;

    // State machine variables
    private StringBuffer clearsignedTextBuffer;
    private Vector payloadVector;
    private CRC accumulatingCRC;

    // The current state of the state machine
    private int state;
    public boolean isComplete() { return (state == COMPLETE); }
    
    // For the "Comment:", etc., parts of the message...
    private Properties properties;
    public Properties getProperties() { return properties; }

    // For the clear text of a clearsigned message
    private byte[] clearText;
    public byte[] getClearText() { return clearText; }

    // The signature, or self-contained armoured packet(s)
    private PacketByteArray payload;
    public PacketByteArray getPayload() { return payload; }



    /**
     * Create an ArmouredData object which is an armour wrapper around
     * any ".pgp" file.
     *
     * @param payload pgp packets to be armoured
     */
    public ArmouredData( PacketByteArray payload )
    {
        this( new Properties(), payload );
    }

    /**
     * Create an ArmouredData object which is an armour wrapper around
     * any ".pgp" file.
     *
     * @param properties properties to appear in armoured message
     *    (eg. "Comment","blah" ).
     *    For each item in the properties, the key <i>must</i> be
     *    in mixed case (first char upper case, rest lower case)
     *    and the value may be in any case.
     * @param payload pgp packets to be armoured
     */
    public ArmouredData( Properties properties, PacketByteArray payload )
    {
        this.properties = properties;
        this.clearText = null;
        this.payload = payload;
        properties.put( "Version", Cryptix.getVersionString() );
        state = COMPLETE;
    }

    /**
     * @deprecated
     */
    public ArmouredData( byte[] text, PacketByteArray signature )
    throws FormatException
    {
        this( new Properties(), text, signature );
    }

    /**
     * @deprecated
     */
    public ArmouredData( Properties properties,
                         byte[] clearText, PacketByteArray signature )
    throws FormatException
    {
        this.properties = properties;
        this.clearText = clearText;
        this.payload = payload;
        properties.put( "Version", Cryptix.getVersionString() );
        if ((payload.numPackets() != 1)
         || (payload.getPacketType( 0 ) != Packet.SIGNATURE))
        {
            throw new FormatException(
                "Payload for clearsigned text is not a Signature packet" );
        }
        state = COMPLETE;
    }

    /**
     * Constructor which does clearsigning.
     * @param plaintext String containing text to be signed.
     * @param key key with which to sign plaintext.
     */
    public ArmouredData( String plaintext, SecretKey key )
        throws FormatException, IOException
    {
        this( new Properties(), plaintext, key );
    }

    /**
     * Constructor which does clearsigning.
     * @param properties properties to appear in armoured message
     *    (eg. "Comment","blah" ).
     *    For each item in the properties, the key <i>must</i> be
     *    in mixed case (first char upper case, rest lower case)
     *    and the value may be in any case.
     * @param plaintext String containing text to be signed.
     * @param key key with which to sign plaintext.
     */
    public ArmouredData( Properties properties, String plaintext, SecretKey key )
        throws FormatException, IOException
    {
        // Convert to canonical linefeeds
        String useText = CRLF.encode( plaintext );
        byte[] useBytes = LegacyString.toByteArray( useText );

        // Convert from local charset
        String charset = properties.getProperty( "Charset", "noconv" );
        byte[] sigBytes = CharSet.decode( useBytes, charset );
        String sigText = LegacyString.toString( sigBytes );

        // Remove trailing spaces
        sigText = removeTrailingSpaces( sigText );
        sigBytes = LegacyString.toByteArray( sigText );

        // Generating signature
        MD5 md = new MD5();
        md.add( sigBytes );
        Signature sig = new Signature( key, md );

        // Convert signature to byte[]
        ByteArrayOutputStream out1 = new ByteArrayOutputStream();
        DataOutputStream out2 = new DataOutputStream( out1 );
        PacketOutputStream out3 = new PacketOutputStream( out2 );
        out3.writePacket( sig );
        out3.flush();
        byte[] buffer = out1.toByteArray();

        // Initialise ArmouredData variables
        properties.put( "Version", Cryptix.getVersionString() );
        this.properties = properties;
        this.clearText = useBytes;
        this.payload = new PacketByteArray( buffer );

        // Set state machine state
        state = COMPLETE;
    }

    /**
     * This constructor accepts text which is already armoured.
     * @param armouredText String containing armoured data
     */
    public ArmouredData( String armouredText ) throws FormatException, DecryptException
    {
        properties = new Properties();
        state = WAITING_FOR_BEGIN;
        addLines( armouredText );
    }

    /**
     * This constructor accepts text which is already armoured.
     * @param armouredText byte array containing armoured data
     */
    public ArmouredData( byte[] armouredBytes ) throws FormatException, DecryptException
    {
        this( LegacyString.toString( armouredBytes ));
    }


    
    /**
     * A constructor for use with the state machine!<p>
     *
     * To use this, you must repeatedly call add() for each line of text,
     * until you run out of lines, and then call isComplete() to see if
     * the state machine agrees with you that the process is complete.<p>
     *
     * @param useStateMachine must be true or else!
     */
    public ArmouredData( boolean useStateMachine ) throws DecryptException
    {
        if (!useStateMachine)
            throw new DecryptException( "Invalid use of state machine constructor" );
        state = WAITING_FOR_BEGIN;
    }

    // Static methods to armour and de-armour a message (EXcluding the
    // "-----BEGIN PGP MESSAGE-----" (etc.), version, etc. wrappers.
    // Variants of these functions were originally placed in the
    // now deprecated Armoury class.

    /**
     * Armour a PGP packet sequence, excluding wrapper.
     * @param pba PacketByteArray to be armoured.
     * @return armoured message, excluding wrapper.
     */
    public static String armour( PacketByteArray pba )
    {
        byte[] mainBytes = pba.toByteArray();
        String mainEncodedString = Base64.encode( mainBytes );

        int crcInt = CRC.checksum( mainBytes );
        byte crcBytes[] = new byte[3];
        crcBytes[0] = (byte)(crcInt >> 16);
        crcBytes[1] = (byte)(crcInt >>  8);
        crcBytes[2] = (byte)(crcInt      );
        String crcEncodedString = Base64.encode( crcBytes );

        return mainEncodedString + "=" + crcEncodedString;
    }

    /**
     * Unarmour an armoured message (excluding wrapper)
     * @param s armoured message excluding wrapper
     * @return dearmoured PGP packet sequence
     */
    public static PacketByteArray unarmour( String s ) throws FormatException
    {
        int crcPos = s.indexOf( "\r\n=" );

        String mainEncodedString = s.substring( 0, crcPos );
        byte[] mainBytes = Base64.decode( mainEncodedString );

        String crcEncodedString = s.substring( crcPos + 3, crcPos + 7 );
        byte[] crcBytes = Base64.decode( crcEncodedString );

        int crcInt = ((crcBytes[0] & 0xFF) << 16)
                   | ((crcBytes[1] & 0xFF) <<  8)
                   | ((crcBytes[2] & 0xFF)      );

        int calculatedCRC = CRC.checksum( mainBytes );
        if (crcInt != calculatedCRC)
        {
            throw new FormatException( "Armoured Data has incorrect CRC" );
        }

        return new PacketByteArray( mainBytes );
    }

    // Convert the properties to a string in PGP's required form
    private String propertiesToString()
    {
        StringBuffer s = new StringBuffer( 256 );
        Enumeration e = properties.propertyNames();
        while (e.hasMoreElements())
        {
            String name  = (String) e.nextElement();
            if (name.length() > 0)
            {
                String value = properties.getProperty( name );
                s.append( name );
                s.append( ": " );

                s.append( value );
                s.append( "\r\n" );
            }
        }
        s.append( "\r\n" );
        return new String( s );
    }


    /**
     * @deprecated Use an approprate constructor instead
     */
    public void parse( String armouredMessage ) throws FormatException, DecryptException
    {
        clearText = null;
        properties = null;
        payload = null;
        state = WAITING_FOR_BEGIN;
        addLines( armouredMessage );
    }

    /**
     * Unarmour one or more lines of an armoured message. The result of the unarmouring
     * accumulates in the invoking object. Use add() if you want finer control.
     *
     * Note that this routine uses the state machine, and therefore lines of
     * armoured text must be presented to this method in the correct order. You should
     * also have used the constructor new ArmouredData( true ) for this method to be of
     * any use.<p>
     *
     * @param armouredMessage a String which consists of a concatenation of lines of text,
     *    each of which is terminated by linebreaks (including the last).
     */
    public void addLines( String armouredMessage ) throws FormatException, DecryptException
    {
        // Convert to canonical linebreaks
        armouredMessage = CRLF.encode( armouredMessage );

        int pos = 0;
        for (;;)
        {
            int nextPos = armouredMessage.indexOf( "\r\n", pos );
            if (nextPos == -1) break;
            add( armouredMessage.substring( pos, nextPos + 2 ));
            pos = nextPos + 2;
        }
        if (!isComplete())
            throw new DecryptException( "Incomplete Armoured Message. State = " + state );
    }


    /**
     * Unarmour a single line of an armoured message. The result of the unarmouring
     * accumulates in the invoking object, but is <i>also</i> returned to you so you can
     * use the dearmoured bytes as they are generated.<p>
     *
     * Note that this routine is the heart of the state machine, and therefore lines of
     * armoured text must be presented to this method in the correct order. You should
     * also have used the constructor new ArmouredData( true ) for this method to be of
     * any use.<p>
     *
     * @param armouredLine a String which consists of a single line of text
     * @return the result of dearmouring this line, or null if this line produced no
     *    new bytes.
     */
    public byte[] add( String armouredLine ) throws DecryptException, FormatException
    {
        // Convert to canonical linebreaks
        armouredLine = CRLF.encode( armouredLine );
        if (!armouredLine.endsWith( "\r\n" )) armouredLine += "\r\n";

        // Undo Quoted Printable format if neccessary.
        while ((armouredLine.indexOf('=') != -1)
            && QuotedPrintable.isProbablyEncoded( armouredLine ))
        {
            armouredLine = QuotedPrintable.decode( armouredLine );
        }

        // Here's the state machine itself
        switch (state)
        {
            case WAITING_FOR_BEGIN:
            {
                // Wait until we reach the first line,
                // then decide what sort of message it is.
                if (armouredLine.equals( "-----BEGIN PGP SIGNED MESSAGE-----\r\n" ))
                {
                    // Next comes clearsigned text
                    clearsignedTextBuffer = new StringBuffer();
                    state = BEGIN_CLEARSIGNED_MESSAGE;
                }
                else if (armouredLine.equals( "-----BEGIN PGP MESSAGE-----\r\n" ))
                {
                    // Go straight to the properties.
                    properties = new Properties();
                    state = COMPILING_PROPERTIES;
                    clearsignedTextBuffer = null;
                }
                break;
            }
            case BEGIN_CLEARSIGNED_MESSAGE:
            {
                // In older versions of PGP, comments were allowed in this field.
                // This has now been disallowed because otherwise text may be inserted
                // which does not affect the signature. Hence, now we allow only a
                // blank line in this field
                if (armouredLine.equals( "\r\n"))
                    state = COMPILING_CLEARSIGNED_MESSAGE;
                else
                    throw new FormatException( "Clearsigned body not preceeded by blank line" );
                break;
            }
            case COMPILING_CLEARSIGNED_MESSAGE:
            {
                if (armouredLine.equals( "-----BEGIN PGP SIGNATURE-----\r\n"))
                {
                    // put the cleartext in place
                    String text = clearsignedTextBuffer.toString();
                    if (!text.endsWith( "\r\n\r\n" ))
                        throw new FormatException( "Clearsigned body not followed by blank line" );

                    text = text.substring( 0, text.length() - 2 );
                    clearText = LegacyString.toByteArray( text );

                    clearsignedTextBuffer = null;
                    state = COMPILING_PROPERTIES;
                }
                else
                {
                    clearsignedTextBuffer.append( armouredLine );
                }
                break;
            }
            case COMPILING_PROPERTIES:
            {
                if (armouredLine.equals( "\r\n" ))
                {
                    state = DEARMOURING;
                    payloadVector = new Vector();
                    accumulatingCRC = new CRC();
                }
                else
                {
                    int pos = armouredLine.indexOf( ": " );
                    if (pos == -1)
                    {
                        throw new FormatException( "Header \"" + armouredLine
                            + "\" does not conform to PGP standard \"Name: Value\"" );
                    }
                    String name  = armouredLine.substring( 0, pos );
                    if (name.length() > 0)
                    {
                        String value = armouredLine.substring( pos+2, armouredLine.length()-2 );
                        properties.put( capitalise( name ), value );
                    }
                }
                break;
            }
            case DEARMOURING:
            {
                if (armouredLine.startsWith( "-----END PGP" ))
                    throw new DecryptException( "Armoured Message has no CRC" );

                if (armouredLine.charAt( 0 ) == '=')
                {
                    state = EXPECTING_END_OF_MESSAGE;
                    String crcEncodedString = armouredLine.substring( 1, 5 );
                    byte[] crcBytes = Base64.decode( crcEncodedString );

                    int crcInt = ((crcBytes[0] & 0xFF) << 16)
                               | ((crcBytes[1] & 0xFF) <<  8)
                               | ((crcBytes[2] & 0xFF)      );

                    if (crcInt != accumulatingCRC.crc())
                    {
                        throw new DecryptException( "Armoured Data has incorrect CRC" );
                    }
                }
                else
                {
                    byte[] dearmouredBytes = Base64.decode( armouredLine );
                    accumulatingCRC.add( dearmouredBytes );
                    payloadVector.addElement( dearmouredBytes );
                    return dearmouredBytes;
                }
                break;
            }
            case EXPECTING_END_OF_MESSAGE:
            {
                if (armouredLine.startsWith( "-----END PGP" ))
                {
                    // put the payload together
                    int n = 0;
                    byte[] buffer;

                    for (int i=0; i<payloadVector.size(); i++)
                    {
                        buffer = (byte[]) payloadVector.elementAt( i );
                        n += buffer.length;
                    }

                    byte[] payloadBuffer = new byte[ n ];
                    n = 0;
                    for (int i=0; i<payloadVector.size(); i++)
                    {
                        buffer = (byte[]) payloadVector.elementAt( i );
                        int len = buffer.length;
                        System.arraycopy( buffer, 0, payloadBuffer, n, len );
                        n += len;
                    }

                    payload = new PacketByteArray( payloadBuffer );
                    payloadVector = null;
                    state = COMPLETE;
                }
                break;
            }
            case COMPLETE:
            {
                break;
            }
            default:
            {
                throw new DecryptException( "Illegal use of state machine" );
            }
        }
        return null;
    }


    /**
     * Armour an unarmoured message
     */
    public String toString()
    {
        StringBuffer sb = new StringBuffer( 8192 );
        if (clearText != null)
        {
            String text = LegacyString.toString( clearText );
            text = dashProtect( text );

            sb.append( "-----BEGIN PGP SIGNED MESSAGE-----\r\n\r\n" );
            sb.append( text );
            sb.append( "\r\n-----BEGIN PGP SIGNATURE-----\r\n" );
            sb.append( propertiesToString() );
            sb.append( armour( payload ));
            sb.append( "-----END PGP SIGNATURE-----\r\n" );
        }
        else
        {
            sb.append( "-----BEGIN PGP MESSAGE-----\r\n" );
            sb.append( propertiesToString() );
            sb.append( armour( payload ));
            sb.append( "-----END PGP MESSAGE-----\r\n" );
        }
        return new String( sb );
    }

    /**
     * Prefix all vulnerable lines with "- ".
     * @param s text to be encoded
     * @return dash-proteced text
     */
    public static String dashProtect( String s )
    {
        int startPos = 0;
        int len = s.length();
        StringBuffer buffer = new StringBuffer( 2 * len );

        while (startPos < len)
        {
            int nextPos = s.indexOf( "\r\n", startPos ) + 2;
            if (nextPos == 1) nextPos = len;

            if ((s.startsWith( "-",     startPos ))
             || (s.startsWith( "From ", startPos ))
             || (s.startsWith( ".\r\n", startPos )))
            {
                buffer.append( "- " );
            }
            buffer.append( s.substring( startPos, nextPos ));
            startPos = nextPos;
        }
        return buffer.toString();
    }

    /**
     * Remove all "- "s from the start of lines.
     * @param s dash-proteced text to be decoded
     * @return text with dash-protection removed
     */
    public static String dashUnprotect( String s )
    {
        int startPos = (s.startsWith( "- " )) ? 2 : 0;
        int len = s.length();
        StringBuffer buffer = new StringBuffer( len );

        while (startPos < len)
        {
            int nextPos = s.indexOf( "\r\n- ", startPos );
            if (nextPos == -1)
            {
                buffer.append( s.substring( startPos ));
                startPos = len;
            }
            else
            {
                buffer.append( s.substring( startPos, nextPos + 2 ));
                startPos = nextPos + 4;
            }
        }
        return buffer.toString();
    }

    /**
     * Remove all trailing spaces from each line,
     *
     * @param s text to be trimmed
     * @return text with trailing spaces removed
     */
    public static String removeTrailingSpaces( String s )
    {
        int len = s.length();
        StringBuffer buffer = new StringBuffer( len );
        int spaceCount = 0;

        for (int i=0; i<len; i++)
        {
            char c = s.charAt( i );
            switch (c)
            {
            case ' ':
                spaceCount++;
                break;
            case '\r':
            case '\n':
                spaceCount = 0;
                buffer.append( c );
                break;
            default:
                for (int j=0; j<spaceCount; j++) buffer.append( ' ' );
                spaceCount = 0;
                buffer.append( c );
                break;
            }
        }
        return buffer.toString();
    }

    /**
     * Convert a string to capitalised lower case,
     *
     * @param s text to be converted
     * @return text converted to capitalised lower case.
     */
    public static String capitalise( String s )
    {
        int len = s.length();
        StringBuffer buffer = new StringBuffer( len );
        buffer.append( Character.toUpperCase( s.charAt( 0 )) );
        for (int i=1; i<len; i++)
        {
            buffer.append( Character.toLowerCase( s.charAt( i )) );
        }
        return buffer.toString();
    }

}
