// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.5.3
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.3        1 July 1997     Ian Brown
//      put in caching and exclusion list functionality
//
// 2.5.2        2 May 1997     Ian Brown
//      tidied up ready for release - doc comments added
//
// 2.5.1        15 Apr 1997     Jill Baker
//      imports changed to new namespace
//
// 2.5.0        24 Mar 1997     Ian Brown
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

package cryptix.pgp;

import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.File;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Properties;
import java.util.Date;

import cryptix.security.rsa.PublicKey;
import cryptix.security.rsa.SecretKey;

/**
 * This class acts as a special type of public keyring which stores
 * keys temporarily as a cache. Keys that are added will expire after
 * the amount of time set by the <code>setMemory()</code> function (by
 * default, forever). KeyCache also maintains a list of keys which
 * either should not be cached (as decided by the user) or do not exist.
 * This is used by <code>KeyClient</code> to maintain a list of users
 * who do not have public keys, and thus save it the time of checking with
 * a keyserver every time their key is requested.<p>
 * The exclusion list also stores the date each excluded key should be removed
 * from that list.
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * @version 2.0, 1/7/1997
 * @author Ian Brown
 */
public final class KeyCache extends PublicKeyRing
{
	/** Stores details of unavailable or excluded keys */
	private Properties excluded;
	
	/** File where this list is stored */
	private File excludeFile;
	
	/** How long (in milliseconds) cache remembers keys and exclusions (0=forever, the default) */
	private long memory = 0;
	
	public KeyCache()
	{
		excluded = new Properties();
		excludeFile = null;
	}
	
	public KeyCache( String filename )
	throws IOException
	{
		super( filename );
		excluded = new Properties();
		excludeFile = null;
	}
	
	/**
	 * Use a user-defined exclusion list
	 */
	public void setExclusionList( Properties exclusions )
	{
		excluded = exclusions;
	}
	
	/**
	 * After calling this function, the properties list will be written
	 * out to this file every time it is altered.
	 */
	public void setExclusionFile( File name )
	throws IOException
	{

		FileInputStream fis = new FileInputStream( name );
		excluded.load( fis );
		fis.close();
		
		excludeFile = name;
	}
	
	/**
	 * @return A list of excluded keys
	 */
	public Properties getExclusionList()
	{
		return excluded;
	}
	
	/**
	 * Records excluded and unavailable keys, and how long they will be so.
	 */
	public void markUnavailable( String keyname )
	throws IOException
	{
		// Does cache remember forever?
		if ( memory == 0 ) excluded.put( keyname, new Date( 0 ).toString() );
		
		// Otherwise, calculate how long the keyname will be excluded for		
		else excluded.put( keyname, new Date( new Date().getTime() + memory ).toString() );
		if ( excludeFile != null )
		{
			// Write change to file
			FileOutputStream fos = new FileOutputStream( excludeFile );
			excluded.save( fos, "Users for whom a key is not available" );
			System.out.println( "Saved exclusion list" );
			fos.close();
		}
	}
	
	/**
	 * Is a key in the exclusion list?
	 */
	public boolean unavailable( String keyname )
	{
		// Need to check if key's exclusion has expired - do that later.
		if ( excluded.getProperty( keyname ) != null ) return true;					
		else return false;
	}
	
	/**
	 * How long should the cache remember keys and exclusions
	 */
	public void setMemory( long milliseconds )
	{
		memory = milliseconds;
	}
}

