// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.5.4
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.4	1 July 1997	Ian Brown
//	added caching of retrieved keys, and exclusion lists
//
// 2.5.3	7 June 1997	Ian Brown
//	added functions to return whole KeyRingEntries as well as PublicKeys
//
// 2.5.2        29 Apr 1997     Ian Brown
//      removed caching for the moment - it doesn't work so was only clutter
//	added doc comments
//
// 2.5.1        15 Apr 1997     Jill Baker
//      imports changed to new namespace
//
// 2.5.0        19 Dec 1996     Ian Brown
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

package cryptix.pgp;

import java.io.*;
import java.net.*;
import java.util.Properties;
import java.util.Date;

import cryptix.security.rsa.PublicKey;
import cryptix.security.rsa.SecretKey;
import cryptix.util.EmailAddress;

/**
 * Base for the classes that fetch keys from keyservers.<p>
 * Next version will cache keys, within and between sessions.<p>
 * WebClient is the only class which currently extends this. Others
 * can be easily written - they simply need to implement the <code>fetch()</code>
 * functions.<p>
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 * @version 1.0 19/12/1996
 * @author Ian Brown
 * @see WebClient
 */
public abstract class KeyClient extends KeyStore
{
	protected String serverName;
	protected int port;	
	
	/** Caches already-fetched keys */
	private KeyCache cache;
	
	
	/**
	 * Set up variables for server connection, and initialise empty
	 * temporary cache and exclusion list. If you want these to be
	 * permanent, use the setCache() and setExclude() functions to
	 * supply your own objects.
	 */
	protected KeyClient( String name, int port_ )
	throws IOException, UnknownHostException
	{
		serverName = name;
		port = port_;
		cache = new KeyCache();
		
	}

	public PublicKey getKey( String userName ) throws IOException
	{
		return fetch( userName ).publicKey();
	}

	public PublicKey getKey( EmailAddress address ) throws IOException
	{
		return fetch( address ).publicKey();
	}

	public PublicKey getKey( EmailAddress address, KeyID keyID ) throws IOException
	{
		return fetch( address, keyID ).publicKey();
	}
	
	public KeyRingEntry getKeyRingEntry( String userName ) throws IOException
	{
		// Try cache first for key
		KeyRingEntry entry = cache.getKeyRingEntry( userName );
		if ( entry != null ) return entry;
		
		// Now check exclusions list
		if ( cache.unavailable( userName ) ) return null;
		
		// OK, so we haven't got or tried to get the key before,
		// so go to the keyserver
		entry = fetch( userName );
		
		if ( entry != null )
		{
			cache.add( entry );
			return entry;
		}
		
		else
		{
			// Put it in the 'not available' list
			cache.markUnavailable( userName );
			return null;			
		}
	}

	public KeyRingEntry getKeyRingEntry( EmailAddress address ) throws IOException
	{		
		// Try cache first for key
		System.out.println( "Trying cache..." );
		KeyRingEntry entry = cache.getKeyRingEntry( address );
		if ( entry != null ) return entry;
		
		// Now check exclusions list
		System.out.println( "Trying exclusion list..." );
		if ( cache.unavailable( address.toString() ) ) return null;
		
		// OK, so we haven't got or tried to get the key before,
		// so go to the keyserver
		System.out.println( "Going to keyserver..." );
		entry = fetch( address );
		
		if ( entry != null )
		{
			cache.add( entry );
			return entry;
		}
		
		else
		{
			// Put it in the 'not available' list
			cache.markUnavailable( address.toString() );
			return null;			
		}
	}

	public KeyRingEntry getKeyRingEntry( EmailAddress address, KeyID keyID ) throws IOException
	{
		// Try cache first for key
		KeyRingEntry entry = cache.getKeyRingEntry( keyID );
		if ( entry != null ) return entry;
		
		// Now check exclusions list
		if ( cache.unavailable( keyID.toString() ) ) return null;
		
		// OK, so we haven't got or tried to get the key before,
		// so go to the keyserver
		entry = fetch( address, keyID );
		
		if ( entry != null )
		{
			cache.add( entry );
			return entry;
		}
		
		else
		{
			// Put it in the 'not available' list
			cache.markUnavailable( keyID.toString() );
			return null;			
		}
	}	

	public KeyRingEntry getKeyRingEntry( KeyID keyID ) throws IOException
	{
		return getKeyRingEntry( null, keyID );
	}
	
	/**
	 * Most subclasses which use a distributed keyserver
	 * system need the key's e-mail address to look it up.
	 * WebClient is the exception, in that it uses a server
	 * that stores everyone's keys.
	 */
	public PublicKey getKey( KeyID keyID ) throws IOException
	{
		return null;
	}

	public String
	getUserName( KeyID keyID )
	{
		return null;
	}

	/**
	 * Allows you to use your own <code>KeyCache</code> object as a cache. If you write this out
	 * to disk when your program quits, then load it back in to a <code>KeyClient</code> object,
	 * you will have a between-sessions cache. By default, KeyClient only caches
	 * keys <i>within</i> sessions.
	 */
	public void
	setCache( KeyCache cache )
	{
		this.cache = cache;
	}
	
	public KeyCache
	getCache()
	{
		return cache;
	}
	
	/**
	 * Has the same effect with the exclusions list as <code>setCache()</code> has with the cache.
	 */
	public void
	setExclusions( Properties exclude )
	{
		cache.setExclusionList( exclude );
	}
	
	public Properties
	getExclusions()
	{
		return cache.getExclusionList();
	}
	
	
	/**
	 * fetch() is overridden by child classes to provide the 'meat' of these classes
	 */
	protected abstract KeyRingEntry fetch( String userName ) throws IOException;
	protected abstract KeyRingEntry fetch( EmailAddress address, KeyID keyID ) throws IOException;
	protected abstract KeyRingEntry fetch( EmailAddress address ) throws IOException;
}
