// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.5.7
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.8    10 Oct 1997     Mike Wynn
//      	readUserIdCertificate() now does not push back a Trust packet
//       when it finds a signature without one.
//
// 2.5.7    27 May 1997     Jill Baker
//      Now closes the underlying stream when it reaches EOF for increased
//      efficiency (see readAnyPacket()). Also, close() method added.
//
// 2.5.6    21 May 1997     Jill Baker
//      Bug fix. readUserIdCertificate() now understands revoked keys.
//
// 2.5.5	10 May 1997	    Ian Brown
//	    Emergency bug fix - can be used to read secret and Web-fetched keys properly now
//
// 2.5.4        06 May 1997     Ian Brown (Jill's idea)
//      Comment packets are now ignored by default.
//	    Use readAnyPacket() if you need to get them.
//	    If you do, you need to check *every* time you read a packet
//	    whether it's a Comment, as they can occur anywhere.
//
// 2.5.3        25 Apr 1997     Jill Baker
//      Added method readNonCommentPacket().
//      Bug fix. Several methods which previously weren't declared as
//      synchronized, are now. (previously only pushback() was synchronized).
//
// 2.5.2        14 Apr 1997     Jill Baker
//      Added method markSupported().
//
// 2.5.1        01 Apr 1997     Jill Baker
//      Simple bug fix. Now better at reading the last key in a secret keyring.
//
// 2.5.0        24 Mar 1997     Ian Brown
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

package cryptix.pgp;

import java.io.*;
import java.util.Vector;

/**
 * Adds readPacket(), readKeyRingEntry() and readUserIdCertificate() functions
 * to DataInputStream.
 * Also allows ONE packet to be pushed back onto the stream.
 * @author Mike Wynn
 * @author Ian Brown
 * @author Jill Baker
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 */
public class PacketInputStream extends DataInputStream
{
	/**@variable store Stores a packet that has been pushed back*/
	protected Packet store = null;
	protected DataInputStream din;
    private boolean closed = false;
    private boolean inClosed = false;

	public PacketInputStream( InputStream input )
	{
		super( input );
		din = new DataInputStream ( in );
	}

	public PacketInputStream( DataInputStream input )
	{
		super( input );
		din = input;
	}

	/**
	 * Reads any packet, including comment packets.
	 */
	public synchronized Packet readAnyPacket() throws IOException
	{
        if (closed) return null;
        Packet temp;
		if ( store == null )
        {
            if (inClosed) return null;
            temp = PacketFactory.read( din );
            if (temp == null)
            {
                in.close();
                inClosed = true;
            }
            return temp;
        }
		else
		{
			temp = store;
			store = null;
			return temp;
		}
	}

	/**
	 * Reads any packet apart from Comment packets.
	 * This greatly simplifies the processing packet streams,
	 * which can contain Comments anywhere.
	 */
	public synchronized Packet readPacket() throws IOException
	{
       		Packet p;
        	do p = readAnyPacket();
        	while ( (p != null) && ( p instanceof Comment ) );
        	return p;
	}

	/**
	 * Push back a packet which has been read onto the stream.
	 * This is essential for parsing multi-packet structures such as
	 * a KeyRingEntry, where you don't know how many packets of what
	 * type to expect.
	 */
	public synchronized void
	pushback( Packet p )
	throws IOException
	{
		if ( store == null ) store = p;
		else throw new IOException( "Only one packet may be pushed back." );
	}

	/**
	 * Used by KeyRing to load in key certificate and accompanying user IDs and signatures
	 * and KeyServer to decipher keys retrieved from remote keyservers.
	 */
	public synchronized KeyRingEntry
	readKeyRingEntry( KeyRingTrust defaultTrust)
	throws IOException
	{
		Packet p;
		Certificate cert;
		UserIdCertificate usercert;
		KeyRingTrust trust;
		Vector userCerts = new Vector();

		// Read certificate packet
		// Need to catch incomplete Entries, when in runs out of data
		p = readPacket();
		if ( p == null ) return null; // Reached the end of the stream
		else if ( ! ( p instanceof Certificate ) ) 
			throw new IOException("Expecting a Certificate in keyring; got a " + p.getClass().getName()  );
		else cert = (Certificate)p;

		// Check for trust packet
		p = readPacket();
		if ( p == null ) return null; // Reached the end of the stream
		else if ( ! ( p instanceof KeyRingTrust ) )
		{
		    // We may be reading a secret keyring, or a keyserver-fetched key entry
		    // which contains no trust information
		    trust = defaultTrust;
		    pushback( p );
		}
		else trust = (KeyRingTrust)p;

		// Read user ID certificates
		while ( ( usercert = readUserIdCertificate( defaultTrust ) ) != null )
			userCerts.addElement( usercert );

		return new KeyRingEntry( cert, trust, userCerts );
	}

	public synchronized UserIdCertificate
	readUserIdCertificate( KeyRingTrust defaultTrust )
	throws IOException
	{
		UserId userid;
		KeyRingTrust trust;
		Vector signatures = new Vector();
		Vector trusts = new Vector();
		Packet p = null;

		// Need to catch incomplete certificates, when in runs out of data
        p = readPacket();

        // Check for revoked key
        if (p instanceof Signature)
        {
            if (((Signature)p).getSignatureType() == Signature.KEY_COMPROMISE_CERTIFICATE)
            {
    			signatures.addElement( p );
                trusts.addElement( defaultTrust );
			// If next line is updated to give translations of KEY REVOKED,
			// remember to update the bit in KeyRingEntry which checks for it
                userid = new UserId( "*** KEY REVOKED ***" );
        		return new UserIdCertificate( userid, defaultTrust, signatures, trusts );
            }
        }

		if ( ! ( p instanceof UserId ) )
		{
			// Reached end of UserIdCertificates section
			if ( p != null ) pushback( p );
			return null;
		}		
		else userid = (UserId)p;
		
		p = readPacket();		
		if ( ! ( p instanceof KeyRingTrust ) )
		{
			// Reached end of UserIdCertificates section
			// or reading from a no-trust-given source
            		trust = defaultTrust;
            		if ( p != null ) pushback( p );
        	}
		else trust = (KeyRingTrust)p;
		
		while ( ( p = readPacket () ) instanceof Signature )
		{
			signatures.addElement( p );
			p = readPacket();
			if ( ! ( p instanceof KeyRingTrust ) )
			{
				if ( p != null ) pushback ( p );
				p = defaultTrust;
			}
			trusts.addElement( p );
		}
		// p will now be the UserId at the start of the next UserIdCertificate
		if ( p != null ) pushback ( p );
		return new UserIdCertificate( userid, trust, signatures, trusts );
	}

    public boolean markSupported()
    {
        return false;
    }

    public void close() throws IOException
    {
        if (!closed)
        {
            store = null;
            in.close();
            closed = true;
            inClosed = true;
        }
    }
}
