package enigma;

import java.io.*;
import java.net.*;

import cryptix.pgp.*;

import cryptix.util.EmailAddress;
import systemics.util.IniFileFormatException;

/**
 * Plugin to handle POP3 connections.
 * @author Ian Brown
 * @version 0.92 Fixed bug which occasionally kept user mailbox locked after an error
 */ 
class POP3 extends Plugin
{
	private static final int DEFAULT_POP_SERVER_PORT = 110;
	private static final int DEFAULT_POP_MY_PORT = 110;
	
        // Initialise variables
        ServerSocket server = null;
        Socket clientSocket = null;
        Socket serverSocket = null;

        DataInputStream clientIn = null;
        DataInputStream serverIn = null;
        PrintStream clientOut = null;
        PrintStream serverOut = null;

        String command = null;
        String reply = null;
    
	public POP3( KeyGlob glob, Passphrase pphrase, Preferences prefs0 )
	throws Exception
	{    	
    		super( glob, pphrase, prefs0 );
    		if ( serverPort == -1 ) serverPort = DEFAULT_POP_SERVER_PORT;
    		if ( myPort == -1 ) myPort = DEFAULT_POP_MY_PORT;
    		
    	}
    
	public void run()
	{
		try
		{
	        	// Initialise socket
        		// Create server object
       			server = new ServerSocket( myPort );
       		}
       		
       		catch ( IOException ioe )
		{
    			if ( verbosity >= NORMAL )
    			{
    				System.err.println( "Couldn't create a TCP socket on port " + myPort );
    				if ( myPort < 1025 ) System.err.println( "You need system administration privileges to use a port lower than 1025." );
    				else System.err.println( "Is another program using it?" );
    				System.err.println( "Try changing the line MyPort = " + myPort + " in the [POP3] section of your " + prefs.name() + " file." );
    			}
    			
    			System.exit( PLUGIN_PROBLEM );
		}
		
		while ( true )
		{	
			try
			{
				// Wait for client to connect then
        			// create socket
        			clientSocket = server.accept();
        			if ( !InetAddress.getLocalHost().equals( clientSocket.getInetAddress() ) && !prefs.acceptRemoteClients() )
					throw new RemoteAccessAttemptException();

        			clientIn = new DataInputStream( clientSocket.getInputStream() );
       				clientOut = new PrintStream( clientSocket.getOutputStream(), true );  // true autoflushes stream after newlines

        			// Open connection to true mail server
        			try
        			{
            				serverSocket = new Socket( InetAddress.getByName( serverName ), serverPort );
        			}

        			catch( UnknownHostException uhe )
        			{	
        				if ( verbosity >= NORMAL )
    					{
    						System.err.println( "Can't find POP3 server " + serverName );
            					System.err.println( "Check that your configuration file " + prefs.name() + " contains the correct Internet name." );       			
        				}
        			}

        			serverIn = new DataInputStream( serverSocket.getInputStream() );
        			serverOut = new PrintStream( serverSocket.getOutputStream(), true );

	
				// Get welcome reply from server
				reply = serverIn.readLine();
				if ( verbosity >= VERBOSE ) System.out.println( reply );
				clientOut.println( reply );

        			do
        			{
           				 // Get command from client and send to server
            				command = clientIn.readLine();
					serverOut.println(command);
            
           				// Get reply
            				reply = serverIn.readLine();      
        
            				// Send it to client
            				clientOut.println(reply);
            
            				// Deal with command
            				// More lines may be on their way from the server...
            
           				if( command.startsWith("RETR") )
            			{
            				StringBuffer PGPmsg = new StringBuffer(5096);
            				boolean gettingPGPmsg = false;
            				boolean gotPGPmsg = false;
						    boolean pgpmime = true;

                				// The *important* one
                				// (TOP retrieves only a certain number of lines;
                				// the result of trying to deal with a part-complete
                				// PGP message could be messy)
             
                				if( reply.startsWith( "+OK" ) )
                				{
								// Send header unprocessed
								do
								{
                       					reply = serverIn.readLine();
										clientOut.println(reply);

										// Is this a PGP/MIME message?
										if (reply.indexOf("protocol=\"application/pgp-encrypted\"") != -1) pgpmime = true;

								} while (reply.length() != 0);
 
								// Send body, checking for PGP data and processing it as appropriate
                    					do
                   					{		
                       						reply = serverIn.readLine();
                       						reply.trim(); // PGP is very fussy about trailing whitespace


                        					// Is reply the start of a PGP message?
                        					if ( reply.startsWith( "-----BEGIN PGP MESSAGE" ) || reply.startsWith( "-----BEGIN PGP SIGNED MESSAGE" ) )
                        					{
                        						gettingPGPmsg = true;
                        						if ( verbosity >= VERBOSE ) System.out.println( "Getting a PGP message..." );
                        					}
                        					
                        					if ( gettingPGPmsg ) 
                        					{
                        						PGPmsg.append(reply).append(Armoury.LINEFEED);
                        						if (reply.startsWith( "-----END PGP" ))
                        						{
                        							// We now have a PGP message in PGPmsg. We need to decrypt it
                        							// and send the plaintext to the client
                        							if ( verbosity >= VERBOSE ) System.out.println( "Decrypting/verifying it..." );
                        							try
                        							{
                        								// Get some timing information
                        								long startTime = System.currentTimeMillis();
                        								ArmouredMessage msg = new ArmouredMessage(PGPmsg.toString(), keyGlob, passphrase);
                        								long elapsedTime = System.currentTimeMillis() - startTime;
                        								if ( verbosity >= VERBOSE )
                        								{
                        									System.out.println( "[POP3] That took " + elapsedTime + " milliseconds." );
                        								 	System.out.println( "Sending to client..." );
                        								}
                        								String msgText = msg.plaintext();
                        								int fakePos = msgText.indexOf( "-The following message is signed-" );
                        								while ( fakePos != -1 )
                        								{
                        									// The decrypted plaintext may be trying to pass itself off as authenticated
                        									// so remove that line 
                        									msgText = msgText.substring( 0, fakePos ) + msgText.substring( msgText.indexOf("\r\n", fakePos)+2 );
                        									fakePos = msgText.indexOf( "-The following message is signed-", fakePos + 32 );
                        								}
                        								if ( msg.authenticated() ) clientOut.println( "-The following message is signed---------------------------------" );
                        								clientOut.println( msgText );
                        								if ( msg.authenticated() )
                        								{
                        									String sender = msg.signedBy();
													        StringBuffer dashes = new StringBuffer(); // Stores the start of the signed by line
													        // Java doesn't have a String * int operator, unfortunately
													        for ( int i = 0; i < ( 64 - sender.length() ); i++ ) dashes.append("-");
													        clientOut.println( dashes.toString() + sender + "-" );
												        }
												        if ( prefs.getBool( "General", "ShowOriginalCiphertext" ) )
												        {
													        clientOut.println( "\nOriginal ciphertext follows:\n" );
													        clientOut.println( msg.ciphertext() );
												        }
                        							}

                        							catch ( Exception e )
                        							{
                        								// If any problems occur, just send the ciphertext to the client
                        								clientOut.println( "A problem occurred during the decryption of the following message." );
                        								clientOut.println( e.getClass().getName() + ": " + e.getMessage() );
                        								clientOut.println();
                        								clientOut.println(PGPmsg.toString());
                        							}
                        		
                        							finally
                        							{
                        								PGPmsg = new StringBuffer(5096);
                        								gettingPGPmsg = false;
                        							}
                        						}
                        					}
                        					
                        					else if ( reply.indexOf( "-The following message is signed-" ) != -1 );
                        					// Someone may be trying to make a plaintext message look as if
                        					// Enigma has authenticated it, so don't pass the text to the mailer
                        					
									// if this is a PGP/MIME message, change the application/octet-stream bodypart to text/plain
									// this is a quick and dirty PGPMIME "fix" -- will add proper, full MIME support later
									else if (reply.startsWith("Content-Type: application/octet-stream"))
										clientOut.println("Content-Type: text/plain");

                        					else clientOut.println( reply ); // Send unencrypted line    
                        			                                          
                    					} while(  !reply.equals(".") ); // CRLF.CRLF is termination signa1
                    					// If a message was damaged or contained only part of a PGP segment, just send the plaintext
                    					if (gettingPGPmsg) clientOut.println(PGPmsg.toString());
						}
                			} 
            
					else if(command.equals("LIST") || command.startsWith("TOP") || command.equals("UIDL"))
					{
                				// Response will be multi-line if +OK

                				if( reply.startsWith("+OK") )
                				{
                    					do
							{
          		           				reply = serverIn.readLine();
								clientOut.println(reply);
							} while (!reply.equals(".") ); // CRLF.CRLF is termination signal
						}
            			}           
				} while ( !command.equals("QUIT") );
			}
        
			catch ( NullPointerException npe )
			{
      			// Some badly-behaved POP3 clients don't close the connection to the server correctly
      			// by sending a QUIT message - they just close the socket. This causes a NullPointerException,
      			// which this catch block deals with by simply restarting the plugin.
      			if ( verbosity >= VERBOSE ) System.err.println( "Mail client or POP3 server broke its connection." );     				
				try
				{
					serverSocket.close();
				}

				catch ( IOException ioe )
				{
				}
      		}
      			
      		catch ( RemoteAccessAttemptException raae )
			{
				if ( verbosity >= NORMAL ) 
				{
					System.err.println( "SECURITY ALERT: someone at Internet address " + clientSocket.getInetAddress() );
					System.err.println( "tried to connect to your POP3 plug-in. This would allow them to decrypt your messages." );
					System.err.println( "The request was denied." );
				}
				
				try
				{
						clientSocket.close();
				}
				
				catch ( IOException ioe )
				{
					if ( verbosity >= NORMAL ) System.err.println( "Stopping Enigma for safety reasons." );
					System.exit( PLUGIN_PROBLEM );
				}
			}
			
      		catch ( Throwable t )
      		{
      			// Problems are 'dealt' with by restarting the program.
      			// This is done in order to allow this plugin to be a thread; to extend Thread you have
      			// to deal with all errors yourself instead of being able to throw some of them to the Java VM.
       			if ( verbosity >= NORMAL ) System.err.println( "Problem in POP3 plug-in: " + t.getMessage() );			
				try
				{
					serverSocket.close();
					clientSocket.close();
				}

				catch ( IOException ioe )
				{
				}      	
			}
      			
      		reset(); // Clear buffers
		}   
	}
    
	public void reset(){} // No state variables to reset here
	
	public String name()
	{
		return "POP3";
	}
   
} // End class POP3
