package enigma;

import cryptix.pgp.*;
import cryptix.security.rsa.SecretKey;

import systemics.util.IniFileFormatException;

/**
 * This class provides a basic building block for protocol handlers to extend.
 * It purposefully leaves most of the details to these subclasses to provide
 * maximum flexibility. It provides type safety and some common variables
 * and function templates for Enigma plug-ins.
 *
 * PluginHelper provides much more help if the protocol works in a similar way
 * to POP3 and SMTP.
 *
 * @author Ian Brown
 * @version 0.9
 */ 
abstract class Plugin extends Thread
{
	protected static final int QUIET = -1;
	protected static final int NORMAL = 0;
	protected static final int VERBOSE = 1;
	
	protected int verbosity = VERBOSE;
	
	protected static final int PLUGIN_PROBLEM = 7; // Error code plugins use when exiting
	
	protected String serverName;
	protected int myPort;
	protected int serverPort;
	
	protected KeyGlob keyGlob;
	protected SecretKey myKey = null;
	
	protected Preferences prefs = null;
	
	protected Passphrase passphrase;
	
	/** The 'meat' of the code, handles messages between client and server */
	abstract public void run();
	
	/** Reset state variables */
	abstract public void reset();
	
	/** A short code for the protocol such as POP3 or SMTP */
	abstract public String name();
	
	protected Plugin( KeyGlob glob, Passphrase pphrase, Preferences prefs0 )
	throws IniFileFormatException, Exception
	{
	    	keyGlob = glob;
    		passphrase = pphrase;
    		prefs = prefs0;

		// Check secret key can be decrypted with pphrase
		try
		{
	 		myKey = keyGlob.getSecretKey( prefs.myName(), passphrase );
		}
 	
 		catch ( DecryptException e )
 		{
 			System.err.println( "Couldn't decrypt secret key for " +  prefs.myName() );
 			System.err.println( "Passphrase may be incorrect." );
 		}
 	
 		if ( myKey == null ) throw new Exception( "Couldn't load your secret key." );
 		
 		serverName = prefs.get( name(), "Server" ); 		
 		if ( serverName == null ) throw new IniFileFormatException( "You MUST specify a server name for " + name() + " in your preferences file." );
 		
 		// Port numbers can be given defaults so aren't critical
 		serverPort = prefs.getInt( name(), "ServerPort" );
 		myPort = prefs.getInt( name(), "MyPort" );
 		
 		verbosity = prefs.verbosity();
 	}
 	
 	 /**
	 * How much reporting each plugin should do to the console.
	 * Can be over-ridden in specific plugins if needed.<p>
	 * Choose from QUIET mode (nothing is reported),
	 * NORMAL (diagnostic information is given for errors)
	 * or VERBOSE (lots of information spewed out; useful
	 * for debugging).<p>
	 * Always use >= when checking verbosity, as other values
	 * may be added such as DEBUG later if necessary.
	 */
 	public void setVerbosity( int level )
 	{
 		switch ( level )
 		{
 			case QUIET: verbosity = QUIET; break;
 			case VERBOSE: verbosity = VERBOSE; break;
 			default: verbosity = NORMAL;
 		}
 	}
}
