package enigma;

import java.io.*;
import java.net.*;

import cryptix.security.rsa.SecretKey;
import cryptix.security.rsa.PublicKey;

import cryptix.pgp.*;

import cryptix.util.EmailAddress;
import systemics.util.IniFileFormatException;

/**
 * Allows protocol handlers to be written that concentrate on the protocol itself,
 * rather than the details of handling communication between client and server.
 * @author Ian Brown
 * @version 0.9
 */
abstract class PluginHelper extends Plugin
{	
        // Initialise variables
        private ServerSocket server = null;
        private Socket clientSocket = null;
        private Socket serverSocket = null;

        private DataInputStream clientIn = null;
        private DataInputStream serverIn = null;
        private PrintStream clientOut = null;
        private PrintStream serverOut = null;

        private String command = null;
        private String reply = null;        
        private String processedCommand = null;
        private String processedReply = null;
	
	protected PluginHelper ( KeyGlob glob, Passphrase pphrase, Preferences prefs0 )
	throws IOException, Exception
	{
		super( glob, pphrase, prefs0 );
	}
	
	/* 
	 * The following functions must be implemented by a plugin.
	 * They allow this class to know when to get more lines from
	 * client and server, when the client or server is about to quit
	 * and which String is used by client or server to terminate
	 * multi-line commands.
	 */
	
	/** Does this command imply the client will send more lines? */
	protected abstract boolean multiLineCommand( String command, String reply );
	
	/** Does this reply imply the server will send more lines? */
	protected abstract boolean multiLineReply( String command, String reply );
	
	/** Which string is used to terminate multi-line data? */
	protected abstract String terminator();
	
	/** Does this command imply the client is about to close its connection? */
	protected abstract boolean quitCommand( String command );
	
	/** Does this reply imply the server is about to close its connection? */
	protected abstract boolean quitReply( String reply );
	
	/*
	 * These commands allow extending plugin code to sit in the middle
	 * of the data stream between client and server and process it as
	 * necessary. Note they do NOT actually communicate with the client
	 * or server, merely tell *this* class what to send to them.
	 */
	 
	/** Send whatever this function returns to the server */
	protected abstract String sendToServer( String command );
	 
	/** Send whatever this function returns to the client */
	protected abstract String sendToClient( String reply );
	
	public void run()
	{
		// Initialise socket
      		// Create server object
      		try
      		{
      			server = new ServerSocket( myPort );
        	}
        	
       		catch ( IOException ioe )
		{
    			System.err.println( "Couldn't create a TCP socket on port " + myPort );
    			if ( myPort < 1025 ) System.err.println( "You need system administration privileges to use a port lower than 1025." );
    			else System.err.println( "Is another program using it?" );
    			System.err.println( "Try changing the line port = " + myPort + " in the [POP3] section of your " + prefs.name() + " file." );
    			System.exit( PLUGIN_PROBLEM );
		}
		
		while ( true )
		{
			try
			{
        			// Wait for client to connect
        			// then create socket
        			clientSocket = server.accept();
        			if ( !InetAddress.getLocalHost().equals( clientSocket.getInetAddress() ) && !prefs.acceptRemoteClients() )
					throw new RemoteAccessAttemptException();

        			clientIn = new DataInputStream(clientSocket.getInputStream());
        			clientOut = new PrintStream(clientSocket.getOutputStream(), true);  // true autoflushes stream after newlines

        			// Open connection to true mail server
        			try
        			{
         				serverSocket = new Socket( InetAddress.getByName( serverName ), serverPort );
        			}

        			catch(UnknownHostException e)
        			{
        				System.err.println( "Can't find your " + name() + " server." );
        				System.err.println( "Check your preferences file  " + prefs.name() + " specifies its name correctly,");
        				System.err.println( "and that your network connection is working.");
        			}

        			serverIn = new DataInputStream( serverSocket.getInputStream() );
        			serverOut = new PrintStream( serverSocket.getOutputStream(), true );

	
				// Get welcome reply from server
				reply = serverIn.readLine();
				clientOut.println( reply );

				do
				{
           				// Get command from client and send to server
            				command = clientIn.readLine();
            				System.out.println("Command: '" + command + "'");
            				serverOut.println( sendToServer( command ) );
            	
            				// Get reply
            				reply = serverIn.readLine();      
        
            				// Send it to client
            				clientOut.println( sendToClient( reply ) );
            				System.out.println("Reply: " + reply);
            
            				// Deal with command
            				// More lines may be on their way from the server...
            
            				if ( multiLineCommand( command, reply ) )
            				{
            					do
            					{
            						command = clientIn.readLine();
            						processedCommand = sendToServer( command );
            						if ( processedCommand != null ) serverOut.println( processedCommand );
            					} while ( !command.equals( terminator() ) );
            				}
            		
            				if ( multiLineReply( command, reply ) )
            				{        
            		       			do
            					{
            						reply = serverIn.readLine();
            						processedReply = sendToClient( reply );
            						if ( processedReply != null ) clientOut.println( processedReply );
            					} while ( !reply.equals( terminator() ) );
            					System.out.println( "Finished sending multi-line reply." );
            				}
  
      
				} while ( !quitCommand( command ) && !quitReply( reply ) );
			}
		
			catch ( NullPointerException npe )
			{
      				// Some badly-behaved clients don't close the connection to the server correctly
      				// by sending a QUIT message - they just close the socket. This causes a NullPointerException,
      				// which this catch block deals with by simply restarting the plugin.
      				
      			}
      			
      			catch ( RemoteAccessAttemptException raae )
			{
				System.err.println( "SECURITY ALERT: someone at Internet address " + clientSocket.getInetAddress() );
				System.err.println( "tried to connect to your " + name() + " plug-in. This would allow them to decrypt your messages." );
				System.err.println( "The request was denied." );
				try
				{
					clientSocket.close();
				}
				
				catch ( IOException ioe )
				{
					// If we can't close the remotely-opened socket, stop the program
					// just in case.
					System.err.println( "Stopping Enigma for safety reasons." );
					System.exit( PLUGIN_PROBLEM );
				}
			}
			
			catch ( RuntimeException rte )
			{
      				// RunTimeExceptions are not usually harmful, so just restart the plugin
      				// after printing a warning. It is important that the plugin runs continually
      				// with as little user intervention as possible.
      				System.err.println( "Small problem with " + name() + " plug-in: " + rte.getMessage() );
			}
      		
      			catch ( Throwable t )
      			{
      				// Any other problems are 'dealt' with by stopping the program.
      				// This is done in order to allow this plugin to be a thread; to extend Thread you have
      				// to deal with all errors yourself instead of being able to throw some of them to the Java VM.
       				System.err.println( "Problem in POP3 plug-in: " + t.getMessage() );
      				System.exit( PLUGIN_PROBLEM );     			
      			}
      			
      			reset(); // Clear buffers
      		}
	}        
}
